//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "IvoryPalette.h"
#include "IvoryDraw.h"
#include "IvoryMiscUtils.h"
//---------------------------------------------------------------------------
#ifdef __IvoryBCB3__
#pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
// internal variables:
static const TIvoryErrorClass iec = iecIvoryPalette;

struct TLogPalette256
{
   WORD palVersion;
   WORD palNumEntries;
   TColor palEntries[256];
};
//---------------------------------------------------------------------------
static inline void ValidCtrCheckPalette(TIvoryPalette*)
{
   new TIvoryPalette(NULL);
}
//---------------------------------------------------------------------------
__fastcall TIvoryPalette::TIvoryPalette(TComponent* AOwner)
 : TIvoryResFileBase(AOwner),
   FDirectDrawPalette(NULL), hPalette(NULL),
   FOriginal0(clBlack), FOriginal255(clWhite),
   FUpdate(true),
   FOnChange(NULL)
{
   ZeroMemory(&FColors, sizeof(FColors));
   FColors[255] = clWhite;
}
//---------------------------------------------------------------------------
__fastcall TIvoryPalette::~TIvoryPalette()
{
   if(DirectDrawPalette)
      TheIvoryDraw->ReleasePalette(this);
}
//---------------------------------------------------------------------------
TColor __fastcall TIvoryPalette::GetColor(int n) const
{
   const char* ivoryfuncname = "GetColor";
   if(n < 0 || n > 255) IVORYERROR("Color out of range", ierOther);
   return FColors[n];
}
//---------------------------------------------------------------------------
TColor __fastcall TIvoryPalette::GetOriginalColor(int n) const
{
   const char* ivoryfuncname = "GetColor";
   if(n < 0 || n > 255) IVORYERROR("Color out of range", ierOther);
   if(n == 0) return FOriginal0;
   else if(n == 255) return FOriginal255;
   else return FColors[n];
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::SetColor(int n, TColor c)
{
   if(n == 0) { FOriginal0 = c; return; }
   else if(n == 255) { FOriginal255 = c; return; }
   const char* ivoryfuncname = "SetColor";
   if(n < 0 || n > 255) IVORYERROR("Color out of range", ierOther);
   FColors[n] = c;
   if(DirectDrawPalette && FUpdate)
   {
      HRESULT ddrval = DirectDrawPalette->SetEntries(0, n, 1, (PALETTEENTRY*)(FColors + n));
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t set color", ierDirectX, ddrval);
      if(OnChange) OnChange(this);
      SetPaletteEntries(hPalette, n, 1, (PALETTEENTRY*)&c);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::SetColors(TColor* c)
{
   TColor* colors = FColors;
   colors++;
   FOriginal0 = *c;
   c++;
   for(int i = 1; i < 255; i++)
      *colors++ = *c++;
   FOriginal255 = *c;
   UpdateAllColors();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::Release()
{
   if(DirectDrawPalette)
   {
      DirectDrawPalette->Release();
      FDirectDrawPalette = NULL;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::UpdateAllColors()
{
   const char* ivoryfuncname = "UpdateAllColors";
   if(DirectDrawPalette && FUpdate)
   {
      HRESULT ddrval = DirectDrawPalette->SetEntries(0, 0, 256, (PALETTEENTRY*)FColors);
      if(ddrval != DD_OK) IVORYDDERROR("Couldn\'t set colors", ierDirectX, ddrval);
      if(OnChange) OnChange(this);
      SetPaletteEntries(hPalette, 1, 254, (PALETTEENTRY*)FColors);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::FastLoadFromBMP(unsigned char* Memory, int offsetpal, int sizepal)
{
   PALETTEENTRY ape[256];
   RGBQUAD* prgb;
   int ncolors = sizepal / sizeof(RGBQUAD);
   Memory += offsetpal;
   prgb = (RGBQUAD*)Memory;
   for(int i = 0; i < ncolors; i++)
   {
      ape[i].peRed   = prgb[i].rgbRed;
      ape[i].peGreen = prgb[i].rgbGreen;
      ape[i].peBlue  = prgb[i].rgbBlue;
      ape[i].peFlags = 0;
   }
   SetColors((TColor*)ape);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::FastLoadFromPCX(unsigned char* Memory, int offsetpal, int sizepal)
{
   PALETTEENTRY ape[256];
   int ncolors = sizepal / 3;
   Memory += offsetpal;
   for(int i = 0; i < ncolors; i++)
   {
      ape[i].peRed   = *Memory++;
      ape[i].peGreen = *Memory++;
      ape[i].peBlue  = *Memory++;
      ape[i].peFlags = 0;
   }
   SetColors((TColor*)ape);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::LoadFromFile()
{
   if(ComponentState.Contains(csDesigning)) return;
   const char* err_msg = "Error loading palette";
   const char* ivoryfuncname = "LoadFromFile";
// if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);
   int FileHandle = NULL;
   unsigned char* Memory = NULL;
   try
   {
      FileHandle = FileOpen(FileName, fmOpenRead);
      if(!FileHandle) IVORYERROR(err_msg, ierOther);
      int Size = FileSeek(FileHandle, 0, 2);
      if(Size < 128) IVORYERROR(err_msg, ierOther);
      Memory = new unsigned char[Size];
      FileSeek(FileHandle, 0, 0);
      FileRead(FileHandle, Memory, Size);
      FileClose(FileHandle); FileHandle = NULL;
      int filetype, width, height, offsetpal, sizepal, offsetpixels, sizepixels;
      ReadFileInfo(Memory, Size, false, filetype, width, height, offsetpal, sizepal, offsetpixels, sizepixels);
      if(filetype <= 0) IVORYERROR(err_msg, ierOther);
      if(Size < offsetpal + sizepal) IVORYERROR(err_msg, ierOther);
      if(filetype == 1) FastLoadFromBMP(Memory, offsetpal, sizepal);
      else if(filetype == 2) FastLoadFromPCX(Memory, offsetpal, sizepal);
   }
   catch(...)
   {
      if(FileHandle) { FileClose(FileHandle); FileHandle = NULL; }
      if(Memory) { delete [] Memory; Memory = NULL; }
      throw;
   }
   if(FileHandle) FileClose(FileHandle);
   if(Memory) delete [] Memory;
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::LoadFromResource()
{
   if(ComponentState.Contains(csDesigning)) return;
   const char* err_msg = "Error loading palette resource";
   const char* ivoryfuncname = "LoadFromResource";
// if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);

   try
   {
      bool BitmapResource = true;
      HRSRC hFoundRes = FindResource(DllHandle, ResourceName.c_str(), RT_BITMAP);
      if(!hFoundRes) { hFoundRes = FindResource(DllHandle, ResourceName.c_str(), RT_RCDATA); BitmapResource = false; }
      if(!hFoundRes) IVORYERROR(err_msg, ierOther);
      HGLOBAL hGlobal = LoadResource(DllHandle, hFoundRes);
      unsigned char* Memory = (unsigned char*)LockResource(hGlobal);
      int Size = SizeofResource(DllHandle, hFoundRes);
      if(!BitmapResource && Size < 128) IVORYERROR(err_msg, ierOther);
      int filetype, width, height, offsetpal, sizepal, offsetpixels, sizepixels;
      ReadFileInfo(Memory, Size, BitmapResource, filetype, width, height, offsetpal, sizepal, offsetpixels, sizepixels);
      if(filetype <= 0) IVORYERROR(err_msg, ierOther);
      if(Size < offsetpal + sizepal) IVORYERROR(err_msg, ierOther);
      if(filetype == 1) FastLoadFromBMP(Memory, offsetpal, sizepal);
      else if(filetype == 2) FastLoadFromPCX(Memory, offsetpal, sizepal);
   }
   catch(...)
   {
      throw;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::Create()
{
   if(DirectDrawPalette == NULL)
   {
      const char* err_msg = "Couldn\'t create palette";
      const char* ivoryfuncname = "Create";
      if(TheIvoryDraw == NULL || TheIvoryDraw->DirectDraw == NULL) IVORYERROR(err_msg, ierIvoryDrawNotCreated);

      IDirectDrawPalette* ddpal;
      HRESULT ddrval = TheIvoryDraw->DirectDraw->CreatePalette(DDPCAPS_8BIT, (PALETTEENTRY*)FColors, &ddpal, NULL);
      if(ddrval != DD_OK) IVORYDDERROR(err_msg, ierDirectX, ddrval);
      FDirectDrawPalette = ddpal;
      if(DirectDrawPalette)
         TheIvoryDraw->AddPalette(this);
      if(OnChange) OnChange(this);
      TLogPalette256 lp;
      lp.palVersion = 0x300;
      lp.palNumEntries = 256;
      CopyMemory(lp.palEntries, FColors, sizeof(TLogPalette256));
//      LOGPALETTE lp1;
//      lp1.palVersion = 0x300;
//      lp1.palNumEntries = 1;
      hPalette = CreatePalette((LOGPALETTE*)&lp);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::BeginUpdate()
{
   FUpdate = false;
}
//---------------------------------------------------------------------------
void __fastcall TIvoryPalette::EndUpdate()
{
   FUpdate = true;
   UpdateAllColors();
}
//---------------------------------------------------------------------------
int __fastcall TIvoryPalette::GetNearestPaletteIndex(TColor c)
{
   return ::GetNearestPaletteIndex(hPalette, c);
}
//---------------------------------------------------------------------------

